package com.revfad.gui;

import com.revfad.search.*;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;

//
// 5/31/1999 fixed a bug in showPage(URL url) (discovered by Dave Langers) that 
// caused the page to be loaded in both
// the target and top (default) windows
// Also, fixed a bug that might cause the progress bar to show 100% (after an errored 
// SearchThread called foundNoMatch(int i)) while files were still being searched.
//
// 6/2/1999 fixed a very stupid bug (discovered by Dave Langers)
// on my part that made it impossible for queries
// containing uppercase characters to ever be found -- the scanned lines were
// converted to lowercase while the keywords/phrases were not
//
// 6/4/1999 renamed to AdvSiteSearcher to differentiate from older SiteSearcher
// Plan to release one more version of SiteSearcher using SearchSieves
//
// 6/9/1999 made many major revisions: renamed SearchThread to DocSearcher because
// it is no longer a subclass of Thread (it instead implements Runnable), polished
// the use of SearchSieves, began tentative support for caching, added the ability
// to demand exact matches, and to ignore text in between lesser than and greater than
// signs (probably HTML)
//
// 6/11/1999 began work on a stripped down version of AdvSiteSearcher
// designed to be much smaller (hopefully a little faster) and to output HTML
// via JavaScript (but without using direct communication; no com.netscape.JSObject needed)
//
// 6/12/1999 renamed it to SearchToHTML...fixed a "bug" that caused the applet to 
// ignore the user defined height and width.
// Also finished converting AdvSiteSearcher to SearchToHTML, uncommented SearchSieve.reset(),
// Renamed the new specialized version of DocSearcher, HDocSearcher...added ability to capture
// the context of a match, the title (of an HTML doc), and the closest anchor to a match.
//
// 7/28/1999 added text changing parameters...mostly for internationalization...
// Also, gave up and cut out the code that was supposed to draw a componentless progressbar.
//
// 4/12/2000 fixed a "Y2K bug reported by several alert users...  I am not sure what 
// I was thinking when I wrote the portion of code calling Date.getYear()... Perhaps that it
// returns the decade? Anyway, in reality, getYear() returns the number of years
// since 1900. Files with modification dates beyond 1999 were listed with dates greater than
// 99 (100 for 2000).
// Note: the whole Date class is deprecated in JDK 1.1
// The code actually changed is found in HDocSearcher.java.
// 
// 4/12/2000 added code that causes the HDocSearcher's runner Thread to wait
// when it is not "doing anything." This should be more efficient than in the
// previous incarnation, where runner would sleep, then periodically wake up to 
// see if there was anything to search.
// 
// 5/3/2000 Patrick Fourneret reported that escape sequences were visible in
// the titles of his search results.  Originally, I actually considered this a
// feature - making sure that the title would display correctly on all browsers.
// After having this "feature" pointed out though, I see that it is actually a 
// problem. The fix was simple - I simply stopped calling makeHTMLSafe on the title.
// I also corrected the spelling of "exclude."  I fixed this several months ago,
// but apparently I ended up switching to an older code base somehow. 
//
// Jul 12 2000 I added a kludgy method to HDocSearcher which will
// finish extracting the title from a document even if a match is
// found within the title. I had been reminded of this behavior several
// times before, but it was Danny Narayan's complaint that spurred me to action.
// See HDocSearcher.finishTitle(StringBuffer).
//
// Jul 12 2000 changed the names of SearchToHTML's methods foundMatch and 
// foundTitle to receiveMatch and receiveTitle respectively. The former names
// seemed unfortunately confusing. Added a new method boolean hasTitle(int index)
// to SearchToHTML which returns whether the title of the document at index 
// has been found.
//
// Jul 12 2000 corrected an embarassing error in ReadMe.html - most of the
// text was recycled from SiteSearcher's ReadMe. Unfortunately, some portions
// of the text that don't apply to SearchToHTML made it through. 
//
// Jul 13 2000 I seem to be writing a lot broken sentences in this bug log.
// But that okay.
// Changed the name of the method "foundNoMatch" to "receiveNoMatch." Again,
// I think that the former name was misleading. Added two new parameters to
// deal with expanded context capabilities: leadingContextLength and
// trailingContextLength - leadingContextLength is very misleadingly named.
// I will probably change it tomorrow. The new parameters I was alluding to
// are "leadingcontextlength" and "trailingcontextlength". Not yet documented! 
// I added a new method to HDocSearcher.java: appendTrailingContext(StringBuffer)
// and changed HDocSearcher's constructor in connection with the new trailing context
// stuff.
//
// Jul 14 2000 Added two new parameters: "xhtml_chkbx_checked" and "exact_chkbx_checked"
// Setting each of these to true will initially "check" the corresponding checkbox
// in the applet's user interface. (This worked well with the LineSearcherApplet.)
// Sorry, I've forgotten who suggested this.
//
// Jul 15 2000 Fixed "bugs" in HDocSearcher.java that would cause an 
// ArrayOutOfBoundsException to be thrown if leadingContextLength==0. Previous to a few
// edits ago, I had required that this value be greather than zero so the code's
// assumption had been a safe one. 
//
// Jul 19 2000 Cleaned up and updated the documentation.
//
// Jul 24 2000 Added a new parameter "max_num_matches" - no more than max_num_matches
// documents will be returned as matches to a search. The default value is the 
// the total number of documents. This parameter was suggested by Danny Narayan.
// Uncovered another bug: boolean searching was actually always false as a search was
// underway because search() called stopAllSearches() _after_ setting searching to true
// and stopAllSearches sets search to false. I should scuttle this code and get on with
// the next generation of applets.
//
// Jul 25 2000 Fixed a bug in appendTrailingContext(StringBuffer). The fix required that
// the method not append directly from the input stream to the context (this was the source
// of the problem) so I renamed appendTrailingContext(StringBuffer) to getTrailingContext().
//
// August 17, 2000 Modified the SearchSieve class - see SearchSieve.java for details.
// Updated and made compliant the ReadMe.
//
// August 22, 2000 Changed the default value of the waitforall parameter from false
// to true. Spruced up the documentation and layout of a few methods.
//
// March 4, 2001 Added 4 new parameters: ignore_tags, exact_matches_only, 
// hide_xhtml_chkbx, and hide_exact_chkbx in response to a request from Frank Verni.
// Slightly cleaned up the implementation of the search method. Added some more JavaDoc
// style comments. Changed a few variable names to hopefully be a little more
// intuitive.
// Version 1.16 = March 4, 2001 release
//
// March 7, 2001. Added a progess bar. Added parameters to deal with progress bar: 
// hide_progress_bar, progress_bar_completed_color, progress_bar_remaining_color,
// progress_bar_text_color, progress_bar_height, progress_bar_width, 
// progress_bar_message, and hide_progress_bar_message. Added parameters: hide_button 
// and hide_searchbox_label.
//
// The UI code has been overhauled -- when all of the elements in a section are
// hidden, that section is not added to the applet. Effectively, the remaining
// sections slide and expand to fill the space. Also, the applet UI elements
// now are justified left rather than center.
// 
// The documentation has been updated to reflect the changes in the applet.
//
// The progress bar seems like a major addition. Utilizing it fully will 
// probably require fiddling with the HTML code used to display the applet.
// Therefore, I've decided to advance the major version number:
// Version 2.0 = March 7, 2001
// ---
//
// March 11, 2001. Major changes have been made to the SearchToHTML and HDocSearcher
// classes. While these changes will hopefully not be noticeable to users of the
// applet (except for perhaps a small increase in speed), they may be very noticeable
// to people who have been working with the Java code. I was planning on advancing the
// major version number every time I introduced a binary incompatibility. I am going to
// hold back on that this time. I did not widely distribute SearchToHTML 2.0, and I
// would rather roll these new changes into the first wide 2.0 release than 
// end up at 3.0 or 4.0. Therefore, I am going to call this day's work SearchToHTML 2.1.
//
// Several methods and class variables were removed from SearchToHTML: addInfo(),
// receiveTitle(), hasTitle(); urls, titles, pageinfo. The information formally
// controlled/held by these has been moved into the HDocSearchers. HDocSearchers now
// pass references to themselves to SearchToHTML (in receiveMatch() and receiveNoMatch())
// rather than an index into the workers array.
// March 11, 2001 = Version 2.1
//
// March 12, 2001. Updated documentation. Added the searchbox_size parameter.
// (It is named searchbox_size to match the name of the property used to
// control the column width of an HTML input element.)
// Version 2.2
//
// March 13, 2001. Tried to make the documentation (ReadMe.html) clearer.
// Cut out code in SearchSieve that would allow any space character to match
// any other space character -- tabs could match line feeds. I guestimated
// that the cost of this feature outweighed the benefits. Also, this behavior
// will not be so easy (or possible) to implement if and when SearchSieves
// are switched to use the Knuth-Morris-Pratt algorithm internally.
// 
// March 17, 2001. Under the MRJ VM, clicking the stop button while a search
// is underway appears to lead to a deadlock. But the debugger shows no deadlocks!
// Added a synchronized block to SearchToHTML's search method. This is a very bad
// thing to do as it might block the AWT event dispatch thread. I am desperate, however.
// [The deadlock problem has apparently been fixed.]
//
// March 19, 2001. Made dramatic changes to the way SearchToHTML searches. HDocSearchers
// are now called SearchDocuments. SearchDocuments do not handle forking off new threads
// to search. Instead, SearchDocument's searchFor method blocks until a search is
// complete. It is the calling code that must take care of threading. Currently,
// SearchToHTML is handling this task. SearchToHTML now implements Runnable. Its run
// method contains the code to dispatch the searches.
//
// March 20, 2001. In the new search code, within the run method, progressBar was
// being used even if displayProgressBar was false -- bad idea -- leading to a
// NullPointerException. Fixed. Added code to allow the Web master to block
// the applet user from searching for certain terms (per a request from Frank Verni).
//
// March 24, 2001. Started working on the documentation.
//
// March 29, 2001. Changed the disallowed terms list to be matched, ignoring
// the case of the terms. This makes this feature much more powerful. Note:
// I intend to tear out this feature from the regular release.
//
// April 4, 2001. Added the start_pattern and stop_pattern parameters.
// The text of a document will not be actively searched until the start_pattern
// has been seen. A search will halt immediately if the stop_pattern is seen.
//
// May 6, 2001. Removed code used to block the user from searching for
// certain terms.
// 
// June 13, 2001. Added code to allow the list of files to be searched to be
// loaded from an external file, as requested by Jon Raclaw.
//
// June 28, 2001. Corrected a serious flaw in the SearchDocument class
// by modifying the CircularStringBuffer class. SearchDocument was
// depending on it being legal to have a zero length buffer.
// CircularStringBuffer would throw an IllegalArgumentException on being
// passed zero for a buffer length. The easiest, and perhaps not too
// terrible, solution was to make it possible to have a zero length
// CircularStringBuffer.
//
// July 4, 2001. Happy Independence Day, USA! SearchDocument (and so SearchToHTML)
// now understands relative paths using the buildURL method. It seems fairly
// obvious that buildURL belongs in a separate class acting as an pseudoextension
// to URL, but, for now, I'm sticking it in SearchDocument. I should probably
// move some of the utility methods from SearchToHTML to SearchDocument to
// reduce coupling.
//
// July 6, 2001. Added support for AND searches -- searches that require
// every term to match before the document is considered to match -- in response
// to a request from Simone Lazzarini. Previously, SearchDocument had only supported 
// OR searches. For an OR search, if a single term matches, the document is
// considered to match.
//
// Added the "match_all_terms" parameter to allow this feature to be turned on
// and off. Set "match_all_terms" to "true" to force the applet to require
// that a document contain every term before that document will be considered to match.
//
// August 11, 2001. Fixed a very serious bug in the CircularStringBuffer class
// that led its toString method to always return "". This led the applet
// to never send the leading context. This problem was pointed out by
// Manuel Allaert.
//
// August 16, 2001. Updated the documentation. Changed the way several color-related
// parameters are loaded to be more intelligible to users.
// 
// August 21, 2001. Added a checkbox to allow the user to control whether
// or not every term is required to match. Fixed a bug with AND searches.
//
// August 23, 2001. Added a new variable to searchresults.html that allows
// one to have the closest anchor included in the main link to a matching doc.
// Updated the documentation.
//
// April 28, 2002. Corrected a few grammatical errors in the bugs log.

/**
 * The SearchToHTML class implements a client-side search engine that 
 * displays its results as HTML. The search results are sent via a query
 * string to an HTML document containing a JavaScript. The JavaScript
 * parses the query string and displays the results.
 * <p>
 *This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 *
 * @author David Faden <a href="mailto:dfaden@gmail.com">dfaden@gmail.com</a>
 * @version SearchToHTML 3.0 September 3, 2004
 */
public class SearchToHTML extends Applet implements Runnable, KeyListener, ActionListener, SearchThreadListener {

/**
 * Versioning information.
 */
private static final String version = "SearchToHTML 3.0 September 3, 2004";

//Note: When a change may cause binary imcompatibility -- a method name
//      is changed, a method added or removed, or a radical change in parameters,
//      I will usually advance the major number and set the minor number to zero. 
//      For every other change, including changes to the documentation, I will 
//      advance the minor version number.

/**
 * A description of SearchToHTML.
 */
private static final String description = 
   "A client-side search engine that displays its search results as HTML.";

/**
 * The author of SearchToHTML.
 */
private static final String author = "David Faden <gilbertnews@gmail.com>";

/**
 * Information about SearchToHTML.
 */
private static String appletInfo = version + System.getProperty("line.separator") + 
                             author + System.getProperty("line.separator") + description;


//BUG: This parameter list is currently incomplete. Does anybody actually use
//this? Should we just eliminate it?
/**
 * Information about SearchToHTML's parameters.
 */
private static String[][] parameterInfo = {
    {"files", "url list", "documents to search"},
    {"target", "string", "target frame for search results"},
    {"leadingcontextlength", "integer", "size of leading context, including match"},
    {"trailingcontextlength", "integer", "size of trailing context"},
    {"max_num_matches", "integer", "maximum number of results"},
    {"resultspage", "url", "search results receiver/displayer"},
    {"stop_btn_txt", "string", "text for \"stop\" button"},
    {"bgcolor", "color", "applet background color"},
    {"fgcolor", "color", "applet foreground color"},
    {"buttonbgcolor", "color", "button background color"},
    {"buttonfgcolor", "color", "button foreground color"},
    {"ignore_tags", "boolean", "ignore tags"},
    {"exact_matches_only", "boolean", "accept exact matches only"},
    {"progress_bar_completed_color", "color", "color for completion"},
    {"progress_bar_remaining_color", "color", "color for remaining"},
    {"progress_bar_text_color", "color", "color for progress bar text"},
    {"progress_bar_message", "string", "progress bar message"},
    {"progress_bar_height", "integer", "height of progress bar"},
    {"hide_progress_bar_message", "boolean", "hide progress bar message"},
    {"progress_bar_width", "integer", "width of progress bar"},
    {"send_context", "boolean", "send context"},
    {"send_titles", "boolean", "send titles"},
    {"send_info", "boolean", "send info"},
    {"send_anchors", "boolean", "send anchors"},
    {"phrase_searches_only", "boolean", "phrase searches only"},
    {"files_file", "string", "file containing list of files to search"},
    {"match_all_terms", "boolean", "must all terms match"},
};

/**
 * The number of documents to search.
 */
private int numberOfDocuments;
 
/**
 * The maximum number of matches that will be reported.
 */
private int maxNumOfMatches;

/**
 * Progress bar indicating what percent of the total number of documents have been searched.
 */
private ProgressBar progressBar;
 
/**
 * The frame in which the search results will be displayed.
 */
private String target;

/**
 * Controls whether a message is displayed in an error situation.
 */
private boolean displayMessage;
 
/**
 * Message to be displayed to let the user know what's happening 
 * before the GUI is finished being set up/can't be set up.
 */
private String message;
 
/**
 * The URL of the document displaying the applet. Used as the base
 * for interpreting relative URLs.
 */
private URL baseUrl;
 
/**
  * The URL of the page that will display the search results.
  */
private URL resultsBaseUrl;

/**
 * Controls whether context information is sent with the results.
 */
private boolean sendContext;

/**
 * Controls whether document titles are sent with the results.
 */
private boolean sendTitles;

/**
 * Controls whether document information is sent with the results.
 */
private boolean sendDocumentInfo;

/**
 * Controls whether anchors are sent with the results.
 */
private boolean sendAnchors;

/**
 * Controls whether the search box text is treated as one phrase.
 */
private boolean phraseSearchesOnly;
 
/**
 * Indicates whether a search is currently underway.
 */
private boolean searching = false;

/**
 * The text for the button to stop a search.
 */
private String stop_btn_txt;

/**
 * The parent frame of this applet.
 */
private Frame frame = null;

/**
 * Thread used to run the run method.
 */
private Thread mainThread = null;

/**
 * The number of threads to create for a search. This corresponds to the maximum
 * possible number of simultaneous connections.
 */
private int numberOfThreads = 4;

/**
 * The workers that will carry out the actual work of searching.
 */
private SearchThread[] threads;

/**
 * Hashtable of options controlling the search. See SearchThread.Option for
 * the available keys.
 */
private Hashtable options;

/**
 * The set of documents to search on the next search.
 */
private Vector documents;

/**
 * The queue of documents waiting to be searched.
 */
private DocumentQueue documentQueue;

/**
 * The matches from the most recent search.
 */
private Vector theMatches;

/**
 * Facilitates easily switching from searching mode to display mode.
 */
private CardLayout cardLayout;

/**
 * The button to click to advance to the next page of results.
 */
private Button nextButton;

/**
 * The button to click to advance to the previous page of results.
 */
private Button previousButton;

/**
 * The number of the page of matches last displayed by the applet.
 */
private int pageNumber;

/**
 * The total number of pages of results.
 */
private int numberOfPages;

/**
 * The number of matches to dispay per HTML page.
 */
private int matchesPerPage;

/**
 * The maximum suggested URL length in transmitting search
 * results. Acts as another bound on the number of results per
 * page with higher priority than <tt>matchesPerPage</tt>.
 * The applet will always transmit at least one result per URL
 * regardless of whether the URL is longer than maxUrlLength or not.
 */
private int maxUrlLength;

/**
 * The string the user searched for.
 */
private String searchString;

/**
 * The URLs corresponding to the pages of results. Page 1
 * is at index 0, 2 at 1, and so on.
 */
private Vector pages;

/**
 * Initialize the applet.
 * <br>
 * Read and parse parameters, give meaningful values
 * to variables, and set up the user interface.
 */
public void init() {
	baseUrl = getDocumentBase();
	String baseUrlString = getParameter("base_url");
	if (baseUrlString != null) {
		try {
			baseUrl = new URL(baseUrlString);
		}
		catch (MalformedURLException mue) {
			mue.printStackTrace();
		}
	}

    searching = true;
  
    options = new Hashtable();
   
    int leadingContextLength = getPositiveParameter("leadingcontextlength", 15);
  
    int trailingContextLength = getPositiveParameter("trailingcontextlength", 0);
      
    options.put(SearchThread.Option.LEADING_CONTEXT_LENGTH, new Integer(leadingContextLength));
    options.put(SearchThread.Option.TRAILING_CONTEXT_LENGTH, new Integer(trailingContextLength));
  
    String files = getFilesToSearch();
  
    if (files == null) {
        //TODO: Signal this to the Web master in very noticeable way.

        files = "";
    }
  
    StringTokenizer st = new StringTokenizer(files, "\n\r \t,", false);
    int numberOfTokens = st.countTokens();

    documents = new Vector();  
    for (int i = 0; i < numberOfTokens; i++) {
        String currentToken = st.nextToken();
        try {
            documents.addElement(new Document(Document.buildUrl(baseUrl, currentToken), null));
        }
        catch (MalformedURLException mue) {
            //TODO: Signal this to the Web master in very noticeable way.
            mue.printStackTrace();
        }
    }
    numberOfDocuments = documents.size();
  
    maxNumOfMatches = getPositiveParameter("max_num_matches", 100000000);
  
    target = getParameter("target", "results");
        
    String resultsPage = getParameter("resultspage", "displaymatches.html");
    try {
        resultsBaseUrl = Document.buildUrl(baseUrl, resultsPage);
    }
    catch (MalformedURLException mue) {
        //TODO: Signal this to the Web master in very noticeable way.
        throw new RuntimeException(mue.toString());
    }
    
    sendContext = getParameter("send_context", true);
    
    if (!sendContext) {
        options.put(SearchThread.Option.LEADING_CONTEXT_LENGTH, new Integer(0));
        options.put(SearchThread.Option.TRAILING_CONTEXT_LENGTH, new Integer(0));
    }
    
    sendTitles = getParameter("send_titles", true);
    sendDocumentInfo = getParameter("send_info", true);
    sendAnchors = getParameter("send_anchors", true);
    phraseSearchesOnly = getParameter("phrase_searches_only", false);
    
    initNormalGui();
}


/**
 * Get a String containing a list of the files to be searched.
 * This method looks at the both the "files" and "files_file"
 * parameters in building its result.
 *
 * @return A String containing a list of the files to be searched.
 */
private String getFilesToSearch() {
    String files = getParameter("files");
    
    String moreFiles;
    String filesFile = getParameter("files_file");
    BufferedReader reader = null;
    if (filesFile != null) {
        try {
            URL filesURL = new URL(baseUrl, filesFile);
            
            StringBuffer moreFilesBuffer = new StringBuffer(128);
            reader = new BufferedReader(new InputStreamReader(filesURL.openStream()));
            String currentLine = reader.readLine();
            
            while (currentLine != null) {
                moreFilesBuffer.append(currentLine);
                moreFilesBuffer.append(' ');
                currentLine = reader.readLine();
            }
            
            moreFiles = moreFilesBuffer.toString();
            
            if (files == null) {
                files = moreFiles;
            }
            else {
                files += ',' + moreFiles;
            }
        }
        catch (MalformedURLException mue) {
            mue.printStackTrace();
        }
        catch (IOException ioe) {
            ioe.printStackTrace();
        }
        finally {
            if (reader != null) {
                try {
                    reader.close();
                }
                catch (IOException ioe1) {
                    ioe1.printStackTrace();
                }
            }
        }
    }
    return files;
}

private static final String SEARCHING_PANEL = "SEARCHING_PANEL";
private static final String RESULTS_PANEL = "RESULTS_PANEL";

private void initNormalGui() {
    cardLayout = new CardLayout();
    setLayout(cardLayout);

    //Get this applet's parent frame.
    /*
    Component parent = getParent();
    while (parent != null) {
        if (parent instanceof Frame) {
            frame = (Frame) parent;
            break;
        }
        parent = parent.getParent();
    }
    */

    addKeyListener(this);
  
    setBackground(getParameter("bgcolor", Color.gray));
    setForeground(getParameter("fgcolor", Color.black));
       
    Panel searchingPanel = new Panel(new BorderLayout());
    Color progressBarCompletedPartColor = getParameter("progress_bar_completed_color", Color.blue);
    Color progressBarRemainingPartColor = getParameter("progress_bar_remaining_color", Color.cyan);
    Color progressBarTextColor = getParameter("progress_bar_text_color", Color.black);
    String progressBarMessage = getParameter("progress_bar_message", "Searching... ");
    
    progressBar = new ProgressBar(numberOfDocuments, progressBarCompletedPartColor, progressBarRemainingPartColor, progressBarTextColor, progressBarMessage);

    Dimension d = getSize();
    int progressBarHeight = getPositiveParameter("progress_bar_height", 20);
    if (progressBarHeight > d.height) {
        progressBarHeight = d.height;
    }
    
    int progressBarWidth = getPositiveParameter("progress_bar_width", d.width - 100);
    if (progressBarWidth > d.width) {
        progressBarWidth = d.width;
    }
    
    progressBar.setSize(progressBarWidth, progressBarHeight);
    
    progressBar.setPercentMessageVisible(!getParameter("hide_progress_bar_message", false));
    
    Font progressBarFont = getFontParameter("progress_bar_font");
    if (progressBarFont != null) {
        progressBar.setFont(progressBarFont);
    }
    
    Panel progressBarPanel = new Panel();
    progressBarPanel.add(progressBar);
    searchingPanel.add(progressBarPanel, BorderLayout.WEST);

    progressBar.setProgressBarVisible(false);

    Color buttonBgcolor = getColor(getParameter("buttonbgcolor"));
    Color buttonFgcolor = getColor(getParameter("buttonfgcolor"));
    
    Font buttonFont = getFontParameter("button_font");
    String stopButtonText = getParameter("stop_btn_txt", "Stop");
    
    Button stopButton =
        SearchToHTML.newButton(stopButtonText, buttonBgcolor, buttonFgcolor, buttonFont);
    stopButton.addActionListener(this);
    Panel stopButtonPanel = new Panel();
    stopButtonPanel.add(stopButton);
    searchingPanel.add(stopButtonPanel, BorderLayout.EAST);
    add(searchingPanel, SEARCHING_PANEL);

    Panel resultsPanel = new Panel(new FlowLayout());
    String labelText =
        getParameter("label_txt", "Results Page:");
    Label label = new Label(labelText);
    Font labelFont = getFontParameter("label_font");
    if (labelFont != null) {
        label.setFont(labelFont);
    }
    resultsPanel.add(label);
    
    String previousButtonText = getParameter("previous_btn_txt", "Previous");
    previousButton =
        SearchToHTML.newButton(previousButtonText, buttonBgcolor, buttonFgcolor, buttonFont);
    previousButton.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
            displayPreviousResultsPage();
        }
    });
    resultsPanel.add(previousButton);
    String nextButtonText = getParameter("next_btn_txt", "Next");
    nextButton =
        SearchToHTML.newButton(nextButtonText, buttonBgcolor, buttonFgcolor, buttonFont);
    nextButton.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
            displayNextResultsPage();
        }
    });
    if (buttonBgcolor != null) {
        nextButton.setBackground(buttonBgcolor);
    }
    if (buttonFgcolor != null) {
        nextButton.setForeground(buttonFgcolor);
    }
    resultsPanel.add(nextButton);
    add(resultsPanel, RESULTS_PANEL);
    validate();
}

/**
 * Construct a new Button with the given label. If bgcolor is not
 * null, set the Button's background color to bgcolor. Similarly
 * for fgcolor and the Button's foreground color. If <tt>font</tt> is not
 * null, set the Button's font to <tt>font</tt>.
 *
 * @param label The text of the Button.
 * @param bgcolor The background color, may be null, for the Button.
 * @param fgcolor The foreground color, may be null, for the Button.
 * @param font The font, may be null, for the Button.
 * @return A new Button with the given characteristics.
 */
private static Button newButton(String label, Color bgcolor, Color fgcolor, Font font) {
    Button result = new Button(label);
    if (bgcolor != null) {
        result.setBackground(bgcolor);
    }
    if (fgcolor != null) {
        result.setForeground(fgcolor);
    }
    if (font != null) {
        result.setFont(font);
    }
    return result;
}
 
/**
 * Return information about SearchToHTML.
 *
 * @return Information about SearchToHTML.
 */
public final String getAppletInfo() {
    return appletInfo;
}
 
/**
 * Return information about SearchToHTML's parameters.
 * 
 * @return Information about SearchToHTML's parameters.
 */
public final String[][] getParameterInfo() {
    return parameterInfo;
}


/**
 * Parameter values specified in the query string of this
 * applet's document base.
 */
private Hashtable queryParams;

public String getParameter(String name) {
    if (queryParams == null) {
        String docBase = getDocumentBase().toString();
        String query;
        int q = docBase.indexOf('?');
        if (q != -1) {
            //TODO: Handle Mac IE and Mozilla's erroneous encoding
            //of the URL and query string.
            query = docBase.substring(q + 1);
        }
        else {
            query = super.getParameter("query");
        }
        queryParams = SearchToHTML.parseQuery(query);
    }
    String value = (String) queryParams.get(name);
    if (value != null) {
        return value;
    }
    else {
        return super.getParameter(name);
    }
}

/**
 * Get a Font parameter. If the parameter is unavailable,
 * return null. See java.awt.Font's getFont for a description
 * of the format the underlying parameter must follow.
 * Roughly, you specify the font by
 * its family, style, and size using a space-separated list.
 * For instance, you could use "Arial BOLD 18". The style
 * and size elements of the list are optional.
 *
 * @param parameterName The name of the parameter to get.
 * @return The corresponding Font or null.
 */
private Font getFontParameter(String parameterName) {
    String fontSpec = getParameter(parameterName);
    if (fontSpec == null) {
        return null;
    }
    
    return Font.decode(fontSpec);
}

/**
 * Get a boolean parameter. If the parameter value is unavailable
 * or invalid, return the given default.
 *
 * @param name The name of the applet parameter.
 * @param defaultValue The boolean value to return by default.
 * @return The boolean value for the given parameter.
 */
private boolean getParameter(String name, boolean defaultValue) {
    boolean result = defaultValue;
    String stringValue = getParameter(name);
    if (stringValue != null) {
        if ("true".equalsIgnoreCase(stringValue) || "on".equalsIgnoreCase(stringValue)) {
            result = true;
        }
        else if ("false".equalsIgnoreCase(stringValue)) {
            result = false;
        }
    }
    return result;
}

/**
 * Return the int value of the parameter named <tt>name</tt>.
 * If the parameter is not available, has a value &lt; 0, or
 * an invalid value, return <tt>defaultValue</tt>.
 *
 * @oaram name The name of the parameter to get.
 * *param defaultValue The defaultValue of the parameter.
 *
 * @return The value of the parameter named <tt>name</tt>.
 */
private int getPositiveParameter(String name, int defaultValue) {
    int result = defaultValue;
    String stringValue = getParameter(name);
    if (stringValue != null) {
        try {
            int value = Integer.parseInt(stringValue);
            if (value >= 0) {
                result = value;
            }
            else {
                System.err.println("Negative value for \'" +
                                    name + "\': " + value);
            }
        }
        catch (NumberFormatException nfe) {
            //XXX! We should signal this exception in
            //an easier to see way.
            System.err.println("Invalid value for \'" +
                                name + "\': " + stringValue);
        }
    }
    return result;
}

/**
 * Return the parameter named <code>name</code> if it's available;
 * otherwise, return <code>defaultValue</code>.
 * 
 * @param name String with the name of the desired parameter.
 * @param defaultValue String containing the value to be returned 
 *      if the requested parameter is unavailable.
 * @return Either the requested parameter or defaultValue.
 */
private final String getParameter(String name, String defaultValue) {
    String value = getParameter(name);
    if (value != null) {
        return value;
    }
    return defaultValue;
}

/**
 * Get a color parameter. If the parameter value is unavailable
 * or invalid, return the given default.
 *
 * @param name The name of the applet parameter.
 * @param defaultValue The Color instance to return by default.
 * @return The Color for the given parameter.
 */
private Color getParameter(String name, Color defaultValue) {
    Color result = defaultValue;
    String stringValue = getParameter(name);
    if (stringValue != null) {
        Color aColor = SearchToHTML.getColor(stringValue);
        if (aColor != null) {
            result = aColor;
        }
    }
    return result;
}

/**
 * Return the color corresponding to the given <code>String s</code>.
 * <br>
 * If <code>s</code> starts with a pound sign ('#'), then
 * the following characters will be treated as a hex color string,
 * as in HTML, except that triples are not supported.
 * <br>
 * Otherwise, first <code>s</code> is compared to a number of provided
 * colors, such as "darkblue"; finally, the applet has the system attempt to
 * look up the color.
 *
 * @param s String specifying the color.
 * @return The color matching <code>s</code> or <code>null</code> if
 *         <code>s</code> is not found.
 */
public static final Color getColor(String s) {
    if (s == null) {
        return null;
    }
    
    if (s.startsWith("#")) {
        if (s.length() != 7) {
            //XXX! Should throw an IllegalArgumentException or some such.
            return null;
        }
        else {
            try {
                //Parse a hex string to decimal:
                int num = Integer.parseInt(s.substring(1, 7), 16); 
                return new Color(num);
            }
            catch (NumberFormatException e) {
                //XXX! Should throw an IllegalArgumentException or some such.
                return null;
            }
        }
    }
    else if ("black".equalsIgnoreCase(s)) 
        return Color.black;
    else if ("blue".equalsIgnoreCase(s)) 
        return Color.blue;
    else if ("darkblue".equalsIgnoreCase(s)) 
        return Color.blue.darker().darker().darker();
    else if ("lightblue".equalsIgnoreCase(s)) 
        return Color.blue.brighter().brighter().brighter();
    else if ("cyan".equalsIgnoreCase(s)) 
        return Color.cyan;
    else if ("darkgray".equalsIgnoreCase(s)) 
        return Color.darkGray;
    else if ("lightgray".equalsIgnoreCase(s)) 
        return Color.lightGray;
    else if ("green".equalsIgnoreCase(s)) 
        return Color.green;
    else if ("gray".equalsIgnoreCase(s)) 
        return Color.gray;
    else if ("magenta".equalsIgnoreCase(s)) 
        return Color.magenta;
    else if ("orange".equalsIgnoreCase(s)) 
        return Color.orange;
    else if ("pink".equalsIgnoreCase(s)) 
        return Color.pink;
    else if ("red".equalsIgnoreCase(s)) 
        return Color.red;
    else if ("white".equalsIgnoreCase(s)) 
        return Color.white;
    else if ("yellow".equalsIgnoreCase(s)) 
        return Color.yellow;
    else 
        return Color.getColor(s);
}

/**
 * Parse out the key-value pairs from the given query.
 * Parsing goes from left to right
 * so that the rightmost value of a key that will be 
 * returned. The keys are assumed to be unescaped
 * while the values are assumed to be in escaped UTF-8.
 * If query begins with a question mark, the question mark
 * is deleted before parsing continues.
 * If query is null, an empty Hashtable is returned.
 *
 * @param query The query String to parse.
 * @return A Hashtable of the key value pairs in url.
 */
public static Hashtable parseQuery(String query) {
    Hashtable result = new Hashtable();
    if (query == null) {
        return result;
    }
    if (query.startsWith("?")) {
        query = query.substring(1);
    }
    int hash = query.indexOf('#');
    if (hash != -1) {
        query = query.substring(0, hash);
    }
    StringTokenizer tokenizer = new StringTokenizer(query, "&");
    while (tokenizer.hasMoreTokens()) {
        String pair = tokenizer.nextToken();
        int i = pair.indexOf('=');
        if (i != -1) {
            String value = decodeEscapedUtf8(pair.substring(i + 1));
            if (value != null) {
                result.put(pair.substring(0, i), value);
            }
        }
    }
    return result;
}

/**
 * Unesape the URL-encoded bytes in <tt>escaped</tt> and decode
 * the UTF-8.
 *
 * @param escaped A String containing URL-escaped UTF-8 char.
 * @return The decoded unescaped String corresponding to encoded.
 * Return null if <tt>escaped</tt> is an invalid URL-escaped sequence
 * or corresponds to an invalid UTF-8 sequence.
 */
public static String decodeEscapedUtf8(String escaped) {
    int escapedLength = escaped.length();
    byte[] unescaped = new byte[escapedLength];
    try {
        int unescapedLength = 0;
        for (int i = 0; i < escapedLength; i++) {
            char c = escaped.charAt(i);
            if (c == '%') {
                int high = hexToInt(escaped.charAt(++i));
                int low = hexToInt(escaped.charAt(++i));
                unescaped[unescapedLength] = (byte) (high << 4 | low);
            }
            else if (c == '+') {
                unescaped[unescapedLength] = (byte) ' ';
            }
            else if (0 <= c && c <= 255) {
                unescaped[unescapedLength] = (byte) c;
            }
            else {
                return null;
            }
            unescapedLength++;
        }
        return new String(unescaped, 0, unescapedLength, "UTF8");
    }
    catch (IllegalArgumentException argumentException) {
        return null;
    }
    catch (IndexOutOfBoundsException boundsException) {
        return null;
    }
    catch (UnsupportedEncodingException encodingException) {
        return null;
    }
}

/**
 * Convert a hexadecimal char to it's numerical value.
 *
 * @param c The char to convert.
 * @return The hex value.
 * @throws IllegalArgumentException if c is not a valid hex digit.
 */
private static final int hexToInt(char c) {
    if (c >= '0' && c <= '9') {
         return c - '0';
    }
    else if (c >= 'a' && c <= 'f') {
        return (c - 'a') + 10;
    }
    else if (c >= 'A' && c <= 'F') {
        return (c - 'A') + 10;
    }
    throw new IllegalArgumentException("Not hex: " + c);
}

/**
 * Return a string that will display the char of the input string <tt>text</tt>
 * when displayed in HTML. Replace lesser-than signs, and greater-than signs with
 * the corresponding HTML escapes and replace multibyte char with HTML hex escapes.
 * If escapeAmpersands is true, escape ampersands as well.
 *
 * @param text The string to represent as HTML.
 * @return The HTML representation of <tt>text</tt>.
 */
private static String escapeForHtml(String text, boolean escapeAmpersands) {
    StringBuffer html = new StringBuffer();
    int length = text.length();
    for (int i = 0; i < length; i++) {
        char c = text.charAt(i);
        if (c > 127) {
            html.append("&#x");
            html.append(Integer.toHexString(c));
            html.append(';');
        }
        else if (c == '<') {
            html.append("&lt;");
        }
        else if (c == '>') {
            html.append("&gt;");
        }
        else if (escapeAmpersands && c == '&') {
            html.append("&amp;");
        }
        else {
            html.append(c);
        }
    }
    return html.toString();
}
 
/**
 * Replace every occurrence of char <tt>out</tt> in <tt>track</tt> with <tt>in</tt>.
 *
 * @param track The String to work with.
 * @param out The char to replace.
 * @param in The String to replace out with.
 * @return A copy of <tt>track</tt> with every occurrence of <tt>out</tt> replaced with <tt>in</tt>.
 */
private static String replaceChar(String track, char out, String in) { 
    
    if (track.indexOf(out) < 0)
        return track;
    
    StringBuffer train = new StringBuffer();
    char c;
    int len = track.length();
    for (int i = 0; i < len; i++) {
        c = track.charAt(i);
        if (c == out) {
            train.append(in);
        }
        else {
            train.append(c);
        }
    }
    return train.toString();
}


public void start() {
    if (mainThread == null) {
        mainThread = new Thread(this, "SearchToHTML main thread");
        mainThread.start();
    }
}

public void paint(Graphics g) {
    if (displayMessage) {
        g.drawString(message, 3, getSize().height/2);
        //Instead of this, we should clear the interface
        //and print out the error in a user-selectable
        //TextArea to facilitate bug reports.
    }
}


public void keyPressed(KeyEvent e) {
}

public void keyReleased(KeyEvent e) {
    
}

public void keyTyped(KeyEvent e) {
    if (e.getKeyChar() == '.' && e.isMetaDown() ||
        e.getKeyCode() == KeyEvent.VK_ESCAPE) {
        cancelSearch();
    }
}

/**
 * Handle a click on the stop button.
 *
 * @param e A click on the stop button
 */
public void actionPerformed(ActionEvent e) {
    cancelSearch();
}

/**
 * Display the previous page of results if any in the browser. If there is
 * no such page, do nothing.
 */
public void displayNextResultsPage() {
    if (pageNumber < numberOfPages - 1) {
        pageNumber++;
        displayResultsPage();
    }
}

/**
 * Display the next page of results if any in the browser. If there is
 * no such page, do nothing.
 */
public void displayPreviousResultsPage() {
    if (pageNumber > 0) {
        pageNumber--;
        displayResultsPage();
    }
}

/**
 * Display the current page of results in the browser.
 * Enable or disable the navigation buttons to reflect
 * whether there is a next or previous page.
 */
public void displayResultsPage() {
    if (pageNumber == 0) {
        previousButton.setEnabled(false);
    }
    else {
        previousButton.setEnabled(true);
    }
    
    if (pageNumber == numberOfPages - 1 || numberOfPages < 2) {
        nextButton.setEnabled(false);
    }
    else {
        nextButton.setEnabled(true);
    }
    URL results = (URL) pages.elementAt(pageNumber);
    getAppletContext().showDocument(results, target);
}

/**
 * Cancel the present search.
 */
private void cancelSearch() {
//TODO: If a "significant number" of search results have been collected,
//pop up a dialog asking the users if they want to display what
//has been found so far.
    searching = false;
    stopSearchThreads();
}

public void run() {
	//TODO: Factor out the common code in the next ten or so lines.
    boolean ignoreTags = getParameter("ignore_tags", false);
    options.put(SearchThread.Option.IGNORE_TAG_CONTENT, (ignoreTags ? Boolean.TRUE : Boolean.FALSE));
    boolean exactMatchesOnly = getParameter("exact_matches_only", false);
    options.put(SearchThread.Option.EXACT_MATCHES_ONLY, (exactMatchesOnly ? Boolean.TRUE : Boolean.FALSE));
    boolean matchAll = getParameter("match_all_terms", false);
    options.put(SearchThread.Option.MUST_MATCH_ALL, (matchAll ? Boolean.TRUE : Boolean.FALSE));
    boolean allowMultiple = getParameter("allow_multiple", false);
    options.put(SearchThread.Option.ALLOW_MULTIPLE_MATCHES, (allowMultiple ? Boolean.TRUE : Boolean.FALSE));
    
    searchString = getParameter("q");
    if (searchString == null) {
        throw new RuntimeException("q is null!");
    }
    
    //Hack to make it easy to get some version info.
    if ("*version*".equals(searchString)) {
        System.out.println(getAppletInfo());
    }
    
    progressBar.setProgressBarVisible(true);
    progressBar.setMaximumValue(numberOfDocuments);
    
    Vector matchesVector = searchFor(searchString, options);
    Match[] matches = new Match[matchesVector.size()];
    matchesVector.copyInto(matches);
    quicksort(matches, 0, matches.length - 1, getParameter("sort_by_date", false));
    
    if (searching) {
        matchesPerPage = getPositiveParameter("matches_per_page", 25);
        maxUrlLength = getPositiveParameter("max_url_length", 2040);
        pages = paginateResults(searchString, options, matches);
        
        numberOfPages = pages.size();
        pageNumber = 0;
        displayResultsPage();
        cardLayout.show(this, RESULTS_PANEL);
    }
}

/**
 * Sort the elements with indeces in the range low to high inclusive.
 * This code is based off the pseudocode given in Cormen's
 * "Introduction to Algorithms," p. 146.
 *
 * @param matches The array to sort.
 * @param low The lower bound.
 * @param high The upper bound.
 * @param byDate Sort by date.
 */
private static void quicksort(Match[] matches, int low, int high, boolean byDate) {
    if (low < high) {
        int q = partition(matches, low, high, byDate);
        quicksort(matches, low, q - 1, byDate);
        quicksort(matches, q + 1, high, byDate);
    }
}

/**
 * Divide the elements of matches[low .. high] into two sequences,
 * all elements in the lower-indeced sequence less than or equal to 
 * the given pivot, all elements in the higher sequence
 * greater than the pivot. Return the sorted position of the pivot element.
 *
 * @param matches The array to partition.
 * @param low The lower index inclusive.
 * @param high The higher index inclusive.
 * @param byDate Sort by date.
 * @return The sorted position of the pivot element.
 */
private static int partition(Match[] matches, int low, int high, boolean byDate) {
    Match m = matches[high];
    int i = low - 1;
    for (int j = low; j < high; j++) {
        Match mj = matches[j];
        if ((byDate && mj.compareDateTo(m) <= 0) ||
            (!byDate && mj.compareTo(m) <= 0)) {
            i++;
            swap(matches, i, j);
        }
    }
    swap(matches, i + 1, high);
    return i + 1;
}

/**
 * Swap the values at array[i] and array[j].
 */
private static void swap(Object[] array, int i, int j) {
    Object oldJ = array[j];
    array[j] = array[i];
    array[i] = oldJ;
}

//XXX! We must be mindful of exceeding the Windows imposed limit
//INTERNET_MAX_URL_LENGTH (around 2048 bytes)
//See http://blogs.msdn.com/oldnewthing/archive/2003/12/10/56028.aspx

/**
 * Construct the URLs corresponding to the search results. Do our best
 * to keep each URL under maxUrlLength. 
 */
public Vector paginateResults(String query, Hashtable options, Match[] matches) {
	String resultsBase = resultsBaseUrl.toString();
	int numberOfMatches = matches.length;
	
	if (numberOfMatches == 0) {
		Vector result = new Vector(1);
		try {
			URL resultUrl = new URL(resultsBase + "?" + queryStringPrefix(1, 1, query, options, numberOfMatches));
			result.addElement(resultUrl);
			return result;
		}
		catch (MalformedURLException mue) {
			throw new RuntimeException(mue.toString());
		}
	}
	
	Vector matchesStrings = new Vector();
	int estimatedPrefixLength = resultsBase.length()
								+ 3 * query.length() //Every char URL-escaped
								+ 8 // Number of pages + page number
								+ (int) Math.log(numberOfMatches + 1)
								+ 6 // Separators and y/n
								+ 1; //Question mark.
	int i = 0;
	while (i < numberOfMatches) {
		StringBuffer matchesBuffer = new StringBuffer();
		int end = Math.min(numberOfMatches, i + matchesPerPage);
		matchesBuffer.append(matchToQueryString(matches[i]));
		int length = estimatedPrefixLength + matchesBuffer.length();
		i++;
		while (i < end) {
			String next = matchToQueryString(matches[i]);
			if (next.length() + length < maxUrlLength) {
				length += next.length();
				matchesBuffer.append(next);
				i++;
			}
			else {
				break;
			}
		}
		matchesStrings.addElement(matchesBuffer.toString());
	}
	int numberOfPages = matchesStrings.size();
	
	Vector result = new Vector(numberOfPages);
	for (int j = 0; j < numberOfPages; j++) {
		try {
			URL resultUrl =
				new URL(resultsBase + "?"
					+ queryStringPrefix(numberOfPages, j + 1,
									query, options, numberOfMatches)
					+ (String) matchesStrings.elementAt(j));
			result.addElement(resultUrl);
		}
		catch (MalformedURLException mue) {
			throw new RuntimeException(mue.toString());
		}
	}
	return result;
}

/**
 * Construct the prefix for the query strings representing search results
 *
 * @param numberOfPages The total number results pages.
 * @param pageNumber The page number of the current page.
 * @param query The string the user searched for.
 * @param options The options used for this search.
 * @param numberOfMatches The number of matches found.
 * @return The prefix representing the given search info.
 */
public String queryStringPrefix(int numberOfPages, int pageNumber, String query, Hashtable options, int numberOfMatches) {
    StringBuffer result = new StringBuffer();
    result.append(numberOfPages);
    result.append(',');
    result.append(pageNumber);
    result.append(',');
    String escapedQuery = URLEncoder.encode(escapeForHtml(query, true));
    result.append(replaceChar(escapedQuery, '+', "%20"));
    result.append(',');
    result.append(numberOfMatches);
    result.append(',');
    if (SearchThread.Option.EXACT_MATCHES_ONLY.getFromWithDefault(options, false)) {
        result.append('y');
    }
    else { 
        result.append('n');
    }
    
    if (SearchThread.Option.IGNORE_TAG_CONTENT.getFromWithDefault(options, false)) {
        result.append('y');
    }
    else {
        result.append('n');
    }
    return result.toString();
}

/**
 * URL-encode the given string, using "%20" for spaces
 * and leaving periods and slashes unescaped.
 *
 * @param text The text to escape.
 * @return The escaped version of text.
 */
private static String urlEncode(String text) {
    StringTokenizer tokenizer = new StringTokenizer(text, " ./", true);
    StringBuffer result = new StringBuffer();
    while (tokenizer.hasMoreTokens()) {
        String next = tokenizer.nextToken();
        if (" ".equals(next)) {
            result.append("%20");
        }
        else if ("/".equals(next) || ".".equals(next)) {
            result.append(next);
        }
        else {
            result.append(URLEncoder.encode(next));
        }
    }
    return result.toString();
}

/**
 * Construct the segment of the query string corresponding to the
 * given Match. The segment has the following format
 *<pre>
 (matchURI) ":" (title)? ":" (anchor)? ":" (fileSize)? ":" (lastModified)? ":" (matchContext)? ","
 *</pre>
 * where fileSize is the size in bytes and lastModified is number
 * of milliseconds since Jan 1, 1970. 
 *
 * @param aMatch The Match to represent.
 * @return The query string representation of aMatch.
 */
public String matchToQueryString(Match aMatch) {
	StringBuffer result = new StringBuffer();
	Document doc = aMatch.document();
	
	result.append(urlEncode(doc.pathRelativeTo(resultsBaseUrl)));
	result.append(':');
	
    String title = doc.title();
	if (title != null && sendTitles) {
		result.append(urlEncode(escapeForHtml(title, false)));
	}
	result.append(':');
	
	String anchor = aMatch.nearestAnchor();
	if (anchor != null && sendAnchors) {
		result.append(urlEncode(anchor));
	}
	result.append(':');
	
	int size = aMatch.document().contentLength();
	if (size > 0 && sendDocumentInfo) {
	   result.append(size);
	}
	result.append(':');
	
	long lastModified = doc.lastModified();
	if (lastModified > 0 && sendDocumentInfo) {
		result.append(lastModified);
	}
	result.append(':');
	
	String context = aMatch.context();
	if (context != null && sendContext) {
		result.append(urlEncode(escapeForHtml(context, true)));
	}
	result.append(',');
	return result.toString();
}

/**
 * Construct the query string (excluding the leading question mark)
 * representing the given search results. The returned String has the following
 * format (subject to change as the applet evolves): 
 * <pre>
 numberOfPages "," pageNumber "," searchPhrase "," numberOfMatches "," 
  ("y" | "n") ("y" | "n") 
 ((matchURI) ":" (title)? ":" (anchor)? ":" (documentInfo)? ":" (matchContext)? ",")*
 * </pre>
 *
 * @param query The query that was searched for.
 * @param options The options used in searching.
 * @param matches The matches resulting from the search.
 * @param pageNumber The page of results to display.
 * @return The query string corresponding to the given set of matches.
 */
/*
public String matchesToQueryString(String query, Hashtable options, Vector matches, int pageNumber) {
    StringBuffer results = new StringBuffer();
    results.append(URLEncoder.encode(escapeForHtml(query, true)));
    results.append(',');
    results.append(matches.size());
    results.append(',');
    results.append(pageNumber);
    results.append(',');
    results.append(numberOfPages);
    results.append(',');
    if (SearchThread.Option.EXACT_MATCHES_ONLY.getFromWithDefault(options, false)) {
        results.append('y');
    }
    else { 
        results.append('n');
    }
    
    if (SearchThread.Option.IGNORE_TAG_CONTENT.getFromWithDefault(options, false)) {
        results.append('y');
    }
    else {
        results.append('n');
    }
    
    int begin = pageNumber * matchesPerPage;
    int end = Math.min(matches.size(), begin + matchesPerPage);
    for (int j = begin; j < end; j++) {
        Match match = (Match)matches.elementAt(j);
        String title = match.document().title();
        String context = match.context();
        String anchor = match.nearestAnchor();
        String info = match.document().documentInfo();
        //XXX! URLEncoder needlessly expands periods and slashes. We should do the encoding
        //ourselves to avoid wasting space.
        results.append(URLEncoder.encode(match.document().pathRelativeTo(resultsBaseUrl)));
        results.append(':');
        if (title != null && sendTitles)
            results.append(URLEncoder.encode(escapeForHtml(title, false)));
        results.append(':');
        if (anchor != null && sendAnchors)
            results.append(URLEncoder.encode(anchor));
        results.append(':');
        if (info != null && sendDocumentInfo)
            results.append(URLEncoder.encode(info));
        results.append(':');
        if (context != null && sendContext)
            results.append(URLEncoder.encode(escapeForHtml(context, true)));
        results.append(',');
    }
    String resultsString = results.toString();
    
    resultsString = replaceChar(resultsString, '+', "%20");
    return resultsString;
}
*/


/**
 * Construct the URL corresponding to the given search results.
 *
 * @param query The query that was searched for.
 * @param options The options used in searching.
 * @param matches The matches resulting from the search.
 * @param The page of results to display.
 * @return The URL corresponding to the given set of matches.
 */
 /*
public URL matchesToUrl(String query, Hashtable options, Vector matches, int pageNumber) {
    String resultsString = matchesToQueryString(query, options, matches, pageNumber);
    try {
        return new URL(resultsBaseUrl + "?" + resultsString);
    }
    catch (MalformedURLException mue) {
        throw new RuntimeException(mue.toString());
    }
}
*/

private int documentsRemainingToFinish;

/**
 * Search for the given query with the given options. Block until the search is completed
 * (or interrupted by another process) and return the Vector of Matches found while searching.
 * Searches are always case insensitive.
 *
 * @param query The query to parse and search for.
 * @param options Hashtable of options controlling the search. See SearchThread.Option for
 * the available keys.
 * @return The Matches found for this search.
 */
public Vector searchFor(String query, Hashtable options) {
    query = query.toLowerCase();
    SearchSieve[] patterns;
    boolean exactMatches = Boolean.TRUE == options.get(SearchThread.Option.EXACT_MATCHES_ONLY);
    if (!phraseSearchesOnly) {
        patterns = SearchSieve.parseQuery(query, exactMatches);
    }
    else {
        patterns = new SearchSieve[]{new SearchSieve(query, exactMatches)};
    }
    
    documentsRemainingToFinish = documents.size();
    documentQueue = new DocumentQueue(documents);
    theMatches = new Vector();
    threads = new SearchThread[numberOfThreads];
    for (int i = 0; i < numberOfThreads; i++) {
        threads[i] = new SearchThread(patterns, documentQueue, this, options);
        threads[i].start();
    }
    
    try {
        for (int i = 0; i < numberOfThreads; i++) {
            threads[i].join();
        }
    }
    catch (InterruptedException ie) {
        ie.printStackTrace();
    }
    
    return theMatches;
}

//SearchThreadListener implementation.

/**
 * Called when a thread starts searching a document. 
 *
 * @param thread The calling SearchThread.
 * @param document The document being searched by thread.
 */
public void searching(SearchThread thread, Document document) {
    getAppletContext().showStatus("Searching " + document.fileName());
}

/**
 * Called when a thread finds a match.
 *
 * @param thread The calling SearchThread.
 * @param match The match that was just found.
 */
public synchronized void foundMatch(SearchThread thread, Match match) {
    if (theMatches.size() >= maxNumOfMatches) {
        stopSearchThreads();
    }
    else {
        theMatches.addElement(match);
        getAppletContext().showStatus("Found match: " + match.document().fileName());
        if (theMatches.size() >= maxNumOfMatches) {
            stopSearchThreads();
        }
   } 
}

/**
 * Called when a thread finishes with a document.
 *
 * @param thread The calling SearchThread.
 * @param document The document that was searched.
 */
public synchronized void finishedSearching(SearchThread thread, Document document) {
    progressBar.increment();
    --documentsRemainingToFinish;
    if (documentsRemainingToFinish == 0) {
        stopSearchThreads();
    }
}

/**
 * Called when a thread encounters a reference to another document
 * as through an href or a frame src.
 *
 * @param thread The calling SearchThread.
 * @param document The document containing the link.
 * @param relativeUrl The text of the reference.
 * @param linkText The text of the link. May be null.
 */
public void foundLink(SearchThread thread, Document document, String relativeUrl, String linkText) {
    throw new RuntimeException("Not implemented yet");
}

/**
 * Stop the search threads.
 */
private void stopSearchThreads() {
    if (documentQueue != null) {
        documentQueue.kill();
    }
    
    if (threads!= null) {
        for (int i = 0; i < numberOfThreads; i++) {
            threads[i].stopSearching();
        }
    }
}

public void stop() {
    searching = false;
    stopSearchThreads();
}
 
}
