package com.revfad.search;

// SearchToHTML copyright (c) 1999-2004 David Faden. All rights reserved.
//
// The applet class files and source code are distributed as linkware.
// If you use this applet or a variant on its code,
// include a link to The Gilbert Post, 
// http://www.geocities.com/gilbertnews/
//
// The code and binaries are distributed without warranty of any kind.
// The Gilbert Post and David Faden take no responsibility
// for anything bad that happens as a result of using this applet.
// Please send reports of problems to gilbertnews@hotmail.com, anyway, though.
//

// Modification log:
//
// 4/12/2000 fixed a "Y2K bug" reported by several alert users...  I am not sure what 
// I was thinking when I wrote the portion of code calling Date.getYear()... Perhaps that it
// returns the decade? Anyway, in reality, getYear() returns the number of years
// since 1900. Files with modification dates beyond 1999 were listed with dates greater than
// 99 (100 for 2000).
// Note: the whole Date class is deprecated in JDK 1.1
// The code actually changed is found in HDocSearcher.java.
// 
// 4/12/2000 added code that causes the HDocSearcher's runner Thread to wait
// when it is not "doing anything." This should be more efficient than in the
// previous incarnation, where runner would sleep, then periodically wake up to 
// see if there was anything to search.
//
// Jul 12 2000 I added a kludgy method to HDocSearcher which will
// finish extracting the title from a document even if a match is
// found within the title. I had been reminded of this behavior several
// times before, but it was Danny Narayan's complaint that spurred me to action.
// See HDocSearcher.finishTitle(StringBuffer).
// Finishing the title exposed another problem: the boolean inTitle was not
// set to false even on finding the end of a title.
//
// Jul 13 2000 I seem to be writing a lot broken sentences in this bug log.
// But that okay.
// Changed the name of the method "foundNoMatch" to "receiveNoMatch." Again,
// I think that the former name was misleading. Added two new parameters to
// deal with expanded context capabilities: leadingContextLength and
// trailingContextLength - leadingContextLength is very misleadingly named.
// I will probably change it tomorrow. The new parameters I was alluding to
// are "leadingcontextlength" and "trailingcontextlength". Not yet documented! 
// I added a new method to HDocSearcher.java: appendTrailingContext(StringBuffer)
// and changed HDocSearcher's constructor in connection with the new trailing context
// stuff.
//
// Jul 15 2000 Fixed "bugs" in HDocSearcher.java that would cause an 
// ArrayOutOfBoundsException to be thrown if leadingContextLength==0. Previous to a few
// edits ago, I had required that this value be greather than zero so the code's
// assumption had been a safe one.
//
// Jul 25 2000 Fixed a bug in appendTrailingContext(StringBuffer). The fix required that
// the method not append directly from the input stream to the context (this was the source
// of the problem) so I renamed appendTrailingContext(StringBuffer) to getTrailingContext().
//
// Aug 17, 2000 Changed code to look directly for EOF and on EOF, to pass 
// '\0' to the SearchSieves rather than (char)-1. 
//
// August 22, 2000 Spruced up the documentation of a few methods.
//
// March 9, 2001 Changed the names of several variables and methods (all private)
// to better reflect my current favored style -- that is first word lower case, 
// following words with initial caps. Edited closeConnection so that bis will
// always be set to null, even on an IOException on closeConnection. Messed around
// with conditional statements in several places to try to test for more common
// cases first.
//
// March 10, 2001 Renamed a few more variables to better reflect their function
// and the preferred naming style. Major changes to the internals of doSearch()
// were made, for the sake of efficiency and clarity.
//
// March 11, 2001. Continued the work. Added many more comments. SearchToHTML
// no longer uses indeces to refer to HDocSearchers. Rather, the HDocSearchers
// now pass a reference to themselves back to their parent. Removed the setErrored()
// method.
// SearchSieve has been changed to take a String in its constructor's arguments
// rather than a char array.
//
// March 13, 2001. Information about the document being searched is pulled
// from the HTTP headers in openConnection(). I've changed this method to
// add file size and or creation date information only if that information
// appears to be valid. If the file size <= 0 or the year is less than 1950,
// that respective portion of the information is discarded. Also, a comma is
// now displayed between the month-date and the year. Sorry to everyone
// outside of the U.S.
//
// March 18, 2001. I am going to make quick and dirty edits to this file.
// Proceed with caution. HDocSearcher has been renamed to SearchDocument.
// SearchDocument does not report matches back to a parent SearchToHTML.
// 
// March 19, 2001. SearchDocument (based on HDocSearcher) no longer implements
// Runnable. The searchFor method now blocks until a search is complete and
// returns true or false depending on whether a search term was found. One
// Thread may interrupt another Thread's search by calling interruptSearch.
// An InterruptedException will be thrown from searchFor or searchFor will
// exit normally.
//
// Changed SearchDocument to use the new CircularStringBuffer
// in collecting the context of a match. This is following the example of 
// Marco Soldati, who sent an improved version of the SearchToHTML applet
// to me several months ago. Before, SearchDocument (formerly HDocSearcher)
// implemented the equivalent of a CircularStringBuffer loose within the code.
// 
// Fixed a long standing bug: if a match was found in the title of a document
// and trailingContextLength > 0, the context would be collected, advancing the
// the text stream, and then an attempt was made to collect the title. Bad. Bad.
//
// Two methods were removed: finishTitle(StringBuffer), getTrailingContext().
//
// April 4, 2001 I added methods that allow a search to be stopped on
// hitting a pattern or started only after seeing a pattern. See: 
// getStartPattern setStartPattern(String) stopUsingStartPattern
// isUsingStartPattern setStopPattern(String) getStopPattern 
// stopUsingStopPattern isUsingStopPattern
// 
// When I added support for start and stop patterns to the searchFor method,
// I also rearranged the innards of that method. The code for completing
// the context and a partially retrieved title has been moved out of the 
// search loop. A couple of returns and continues were cut out -- I hope
// this will make it easier for people to follow what is going on in the code.
//
// June 28, 2001. Corrected a serious flaw in the SearchDocument class
// by modifying the CircularStringBuffer class. This bug was pointed out by
// Frank Verni. SearchDocument was depending on it being legal to have a zero 
// length buffer. CircularStringBuffer would throw an IllegalArgumentException 
// on being passed zero for a buffer length. The easiest, and perhaps not too
// terrible, solution was to make it possible to have a zero length
// CircularStringBuffer.
//
// July 4, 2001. Happy Independence Day, USA! SearchDocument (and so SearchToHTML)
// now understands Unix-style relative paths using the buildURL method. It seems fairly
// obvious that buildURL belongs in a separate class acting as an pseudoextension
// to URL, but, for now, I'm sticking it in SearchDocument. I should probably
// move some of the utility methods from SearchToHTML to SearchDocument to
// reduce coupling. It was Scott Beach's request that prompted this feature to
// finally be added.
//
// July 6, 2001. Added support for AND searches -- searches that require
// every term to match before the document is considered to match -- in response
// to a request from Simone Lazzarini. Previously, SearchDocument had only supported 
// OR searches. For an OR search, if a single term matches, the document is
// considered to match.
//
// August 21, 2001. Fixed a bug in the implementation of AND searches. Before, under an
// AND search, some terms would be skipped. The fact that this wasn't caught sooner
// is explained by how hard it was formerly to do AND searches.
//

// June 17 - 18, 2004. Made major changes to work with the new Searcher class
// and to make testing practical.
//
// June 22, 2004 Renamed this class to Document to reflect its generic nature.

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * I represent a searchable document.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 *
 * @author David Faden, dfaden@gmail.com
 * @version June 22, 2004
 */
public final class Document {

/**
 * True if information about the document to be searched has been collected.
 */ 
private boolean gotDocumentInfo;

/**
 * The content type of this document.
 */
private String contentType;

/**
 * The encoding of this document.
 */
private String contentEncoding;

/**
 * The last modification date of this document.
 */
private long lastModified;

/**
 * The content length (in bytes) of this document.
 */
private int contentLength;

/**
 * The URL of this document.
 */
private URL url;

/**
 * The file name of this document, eg, "ex.html".
 */
private String fileName;

/**
 * True if the title of the document has been parsed.
 */
private boolean gotTitle;

/**
 * The title of the document being searched.
 */
private String title;

/**
 * String containing information about the document being searched.
 */
private String documentInfo;
  
/**
 * Construct a new Document. If <tt>encoding</tt> is null,
 * the document will try to guess its encoding.
 * <br>
 * @param url URL of the document to search.
 * @param encoding The text encoding used by this document.
 */
public Document(URL url, String encoding) {
    this.url = url;
    contentEncoding = encoding;
    contentLength = -1;
}


/**
 * Mappings from HTML charset names to Java charset names.
 */
private static Hashtable encodings;

static {

encodings = new Hashtable();
encodings.put("iso-8859-1", "ISO8859_1");
encodings.put("iso8859_1", "ISO8859_1");

encodings.put("macintosh", "MacRoman");

encodings.put("macroman", "MacRoman");

encodings.put("gb2312", "EUC_CN");

encodings.put("euc_cn", "EUC_CN");

encodings.put("big5", "Big5");

encodings.put("utf-8", "UTF8");
encodings.put("utf8", "UTF8");

encodings.put("utf-16", "UTF-16");
encodings.put("utf16", "UTF-16");

//Russian
encodings.put("windows-1251", "Cp1251");
encodings.put("cp1251", "Cp1251");

encodings.put("windows-1252", "Cp1252");
encodings.put("cp1252", "Cp1252");

encodings.put("cp1256", "Cp1256");

encodings.put("big5-hkscs", "Big5_HKSCS");
encodings.put("big5_hkscs", "Big5_HKSCS");

encodings.put("ibm037", "Cp037");
encodings.put("cp037", "Cp037");
encodings.put("ebcdic-cp-us", "Cp037");
encodings.put("ebcdic-cp-ca", "Cp037");
encodings.put("ebcdic-cp-wt", "Cp037");
encodings.put("ebcdic-cp-nl", "Cp037");
encodings.put("csibm037", "Cp037");

encodings.put("ibm273", "Cp273");
encodings.put("cp273", "Cp273");
encodings.put("csIBM273", "Cp273");

//Czech
encodings.put("iso-8859-2", "ISO8859_2");
encodings.put("iso8859_2", "ISO8859_2");

//Russian
encodings.put("iso-8859-5", "ISO8859_5");
encodings.put("iso8859_5", "ISO8859_5");

encodings.put("euc-jp", "JISAutoDetect");
encodings.put("iso-2022-jp", "JISAutoDetect");
encodings.put("shift_jis", "JISAutoDetect");

//Turkish
encodings.put("ibm857", "Cp857");
encodings.put("cp857", "Cp857");
encodings.put("windows-1254", "Cp1254");
encodings.put("cp1254", "Cp1254");
}

private static String normalizedEncoding(String name) {
    return (String) encodings.get(name.toLowerCase());
}

/**
 * Interpret the relative path in <code>relativePath</code>
 * with respect to the base URL <code>base</code> to create
 * a new URL.
 * <br>
 * This method "properly" handles query strings and Unix style relative paths:
 * <ol>
 *   <li>../ (one directory up)
 *   <li>./ (present directory)
 *   <li>/ (root/top directory)
 *</ol>
 *
 * @param base The base URL.
 * @param relativePath String containing a file path relative to base.
 * @return A new URL relative to <code>base</code> according to <code>relativePath</code>.
 * @throws MalformedURLException if relativePath contains enough "../"'s to pop
 * above the root directory level or contains something to confound the URL
 * constructor.
 */
public static URL buildUrl(URL base, String relativePath) 
                                    throws MalformedURLException {
    URL u = null;
    int relPathLength = relativePath.length();
    if (relativePath.startsWith("../")) {
        int relStartIndex = 3; //Index after last ../
        String path = base.getFile();
        
        int dirIndex = path.lastIndexOf('/');
        
        if (dirIndex != -1) {
            dirIndex = path.lastIndexOf('/', dirIndex - 1);
        }
        
        while (relativePath.startsWith("../", relStartIndex)) {
            dirIndex = path.lastIndexOf('/', dirIndex - 1);
            relStartIndex += 3;
            if (dirIndex == -1)
                throw new MalformedURLException("Higher than root: " + base.toString() +
                                                " + " + relativePath);
        }
        
        if (relStartIndex != relPathLength)
            relativePath = relativePath.substring(relStartIndex, relPathLength);
        else
            relativePath = "";
            
        //If the relativeURL hasn't climbed all the way to the root
        //directory, cut the base path to match the directory level
        //calculated in the above code.
        //Otherwise, simply set the path equal to the relativePath
        //with a slash appended to the front to indicate that the
        //relativePath has climbed high enough to become the root
        //directory in the new URL.
        if (dirIndex != -1) {
            path = path.substring(0, dirIndex + 1); //Leave the trailing slash
            if (relStartIndex != relPathLength)
                path += relativePath;
        }
        else
            path = "/" + relativePath;
            
        u = new URL(base.getProtocol(), base.getHost(), path);
    }
    else if (relativePath.startsWith("/")) {
        u = new URL(base.getProtocol(), base.getHost(), 
                            relativePath);
    }
    else if (relativePath.startsWith("./")) {
        u = new URL(base, relativePath.substring(2, relPathLength));
    }
    else if (relativePath.startsWith("?")) {
        u = new URL(base.getProtocol(), base.getHost(), base.getFile() + relativePath);
    }
    else {
        u = new URL(base, relativePath);
    }
    return u;
}

/**
 * Build an absolute URL using this document's URL as
 * the base URL.
 *
 * @param relativeUrl String representing the relative path.
 * @return An absolute URL.
 * @throws MalformedURLException if relativePath contains enough "../"'s to pop
 * above the root directory level or contains something to confound the URL
 * constructor.
 */
public URL buildUrl(String relativeUrl) throws MalformedURLException {
    return Document.buildUrl(this.url(), relativeUrl);
}

public URL url() {
    return url;
}

/**
 * Get the path of the given URL. The URL.getPath
 * method was added after JDK 1.1.8. We reimplement
 * it here for backwards compatibility.
 *
 * @url The URL with the path to get.
 * @return The path component of the given URL.
 */
private static String getPath(URL url) {
    String path = url.getFile();
    int q = path.indexOf('?');
    if (q != -1) {
        path = path.substring(0, q);
    }
    int hash = path.indexOf('#');
    if (hash != -1) {
        path = path.substring(0, hash);
    }
    return path;
}

/**
 * The URL of this document relative to the given
 * url.
 *
 * @param baseUrl The URL to use as a base.
 * @return A relative URL to this document from baseURL.
 * @throws IllegalArgumentException if baseUrl doesn't have
 * the same host as this document.
 */
public String pathRelativeTo(URL baseUrl) {
    if (!baseUrl.getHost().equals(url.getHost())) {
        throw new IllegalArgumentException("Unequal hosts: url.host = " + url.getHost() + " baseUrl.host = " + baseUrl.getHost());
    }
    
    //TODO: Do more extensive compatibiliy checking.
    //For now, just move straight onto the paths.
    
    StringTokenizer basePath = new StringTokenizer(Document.getPath(baseUrl), "/");
    StringTokenizer myPath = new StringTokenizer(Document.getPath(url), "/");
    
    String myToken = null;
    String baseToken = null;
    int relativeDepth = 0;
    boolean atRoot = true;
    
    //First, find the first place where the two paths
    //disagree.
    while (basePath.hasMoreTokens() && myPath.hasMoreTokens()) {
        myToken = myPath.nextToken();
        baseToken = basePath.nextToken();
        if (! myToken.equals(baseToken)) {
            relativeDepth = 1;
            break;
        }
        atRoot = false;
    }
    
    //Climb down the rest of the base hierarchy.
    //We do it this way instead of simply counting the tokens
    //because the last may path component may be a file
    //name rather than a directory.
    //
    //XXX! Would it be better to count the tokens but also
    //check to see if path.indexOf('.') > path.indexOf('/')?
    while (basePath.hasMoreTokens()) {
        baseToken = basePath.nextToken();
        relativeDepth++;
    }
    
    if (baseToken != null
        && baseToken.indexOf('.') != -1
        && !getPath(baseUrl).endsWith("/")) {
        relativeDepth--;
    }
    
    StringBuffer relativePath = new StringBuffer();
    if (atRoot) {
        relativePath.append('/');
    }
    else {
        for (int i = 0; i < relativeDepth; i++) {
            relativePath.append("../");
        }
    }
    
    boolean needSlash = false;
    if (myToken != null) {
        relativePath.append(myToken);
        needSlash = true;
    }
    
    while (myPath.hasMoreTokens()) {
        if (needSlash) {
            relativePath.append('/');
        }
        relativePath.append(myPath.nextToken());
        needSlash = true;
    }
    
    return relativePath.toString();
}

/**
 * The file name of this document. That is, the last component
 * of the path to the document. For example, for
 * http://www.a.com/b/c/d.html this method will return d.html.
 *
 * @return The file name of this document.
 */
public String fileName() {
    if (fileName == null) {
        String file = url().getFile();
        int slashIndex = file.lastIndexOf('/');
        if (slashIndex != -1) {
            fileName = file.substring(slashIndex + 1);
        }
        else {
            fileName = file;
        }
    }
    return fileName;
}

/**
 * Return the title of this document. This may be null.
 *
 * @return The title of this document.
 */
public String title() {
    return title;
}

/**
 * Return the title of this document if the title
 * is not null. Otherwise, return the file name.
 *
 * @return The title of this document if the title
 * is not null. Otherwise, return the file name.
 */
public String titleOrFileNameIfNull() {
    if (title != null) {
        return title;
    }
    else {
        return fileName();
    }
}

private static String[] MONTHS = new String[]{"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

/**
 * A string with info on this document including its
 * last modification date and file size. This method
 * is not interationalization friendly and will be
 * eliminated in a future version.
 *
 * @return A string with info on this document.
 */
/*
public String documentInfo() {
    if (documentInfo == null) {
        Date d = new Date(lastModified());
        int yearsSince1900 = d.getYear();
        int length = contentLength();
        StringBuffer documentInfoBuffer = new StringBuffer();
        
        if (length > 1024) {
            documentInfoBuffer.append(length/1024);
            documentInfoBuffer.append('k');
        }
        else if (length > 0) {
            documentInfoBuffer.append(length);
            documentInfoBuffer.append('b');
        }
        
        if (yearsSince1900 > 71) { //1971
            documentInfoBuffer.append(", ");
            documentInfoBuffer.append(MONTHS[d.getMonth()]);
            documentInfoBuffer.append(' ');
            documentInfoBuffer.append(d.getDate());
            documentInfoBuffer.append(", ");
            documentInfoBuffer.append(yearsSince1900 + 1900);
        }
        
        documentInfo = documentInfoBuffer.toString();
    }
    return documentInfo;
}
*/

public long lastModified() {
    return lastModified;
}

public int contentLength() {
    return contentLength;
}

public String contentType() {
    return contentType;
}

public String contentEncoding() {
    return contentEncoding;
}

/**
 * Open the URL to search and gather some info on it.
 * <br>
 * Pull a little information about the file from
 * the HTTP headers.
 *
 * @return A stream for this document's URL.
 */
private Reader openConnection() throws IOException, SecurityException
{
    URLConnection connection = url.openConnection();
    if (!gotDocumentInfo) {
        gotDocumentInfo = true;
        lastModified = connection.getLastModified();
        if (contentEncoding == null) {
            contentEncoding = connection.getContentEncoding();
            if (contentEncoding != null) {
                contentEncoding = (String) Document.normalizedEncoding(contentEncoding);
            }
        }
        contentType = connection.getContentType();
        contentLength = connection.getContentLength();
    }
    
    if (contentEncoding == null) {
        return guessEncoding(connection.getInputStream());
    }
    else {
        InputStream input = connection.getInputStream();
        try {
            return new BufferedReader(new InputStreamReader(input, contentEncoding));
        }
        catch (UnsupportedEncodingException uee) {
            uee.printStackTrace();
            return new BufferedReader(new InputStreamReader(input));
        }
    }
}

private static final int ENCODING_BUFFER_LENGTH = 400;

private static void copyByteToChar(byte[] byteArray, int length, char[] charArray) {
    for (int i = 0; i < length; i++) {
        charArray[i] = (char) (0xff & byteArray[i]);
    }
}

/**
 * Read from <tt>inputStream</tt> and make a guess at the encoding
 * of the underlying document. The returned Reader is set to start
 * at the same position in the content as <tt>inputStream</tt> was
 * at when it was passed to the method. As a side effect, if a good
 * guess is found, record the document encoding.
 *
 * @param inputStream Stream whose encoding is to be guessed.
 * @return A reader wrapped around the contents of inputStream.
 */
private Reader guessEncoding(InputStream inputStream) throws IOException {
    BufferedInputStream bis = new BufferedInputStream(inputStream);
    final int length = 400;
    byte[] byteBuffer = new byte[ENCODING_BUFFER_LENGTH];
    char[] charBuffer = new char[ENCODING_BUFFER_LENGTH];
    bis.mark(ENCODING_BUFFER_LENGTH);
    //XXX! We need to avoid duplicating the work of finding the shared portions of these patterns.
    //(The second pattern is for matching the output of Microsoft Word. Maybe we should simply reject
    //the second variant?
    SearchSieve charsetFinder1 = new SearchSieve("<meta http-equiv=\"content-type\" content=\"text/html; charset=");
    SearchSieve charsetFinder2 = new SearchSieve("<meta http-equiv=content-type content=\"text/html; charset=");
    int total = 0;
    int numberRead = 0;
    int match = -1;
    while (total < ENCODING_BUFFER_LENGTH && numberRead != -1 && match == -1) {
        numberRead = bis.read(byteBuffer, 0, length - total);
        if (numberRead != -1) {
            total += numberRead;
            copyByteToChar(byteBuffer, numberRead, charBuffer);
            match = charsetFinder1.scanLowerCase(charBuffer, 0, numberRead);
            if (match == -1) {
                match = charsetFinder2.scanLowerCase(charBuffer, 0, numberRead);
            }
        }
    }
    
    if (match != -1) {    
        contentEncoding = finishReadingEncoding(charBuffer, match + 1, numberRead, bis, total);
        contentEncoding = (String) Document.normalizedEncoding(contentEncoding);     
    }
    
    bis.reset();
    if (contentEncoding != null) {
        try {
            return new InputStreamReader(bis, contentEncoding);
        }
        catch (UnsupportedEncodingException uee) {
            uee.printStackTrace();
            //XXX! Should we use 8859-1 here as our default?
            return new InputStreamReader(bis);
        }
    }
    else {
        return new InputStreamReader(bis);
    }
}

/**
 * Finish reading the encoding name.
 *
 * @buffer A char array containing part of the name.
 * @offset The offset into buffer at which the name begins.
 * @numberRead The number of char last read into buffer.
 * @input The stream from which to read any more bytes if necessary.
 * @total The total number of bytes read so far.
 *
 * @return As much of the encoding name as we could safely get.
 */
private String finishReadingEncoding(char[] buffer, int offset, int numberRead, InputStream input, int total) throws IOException {
    int quote = -1;
    for (int i = offset; i < numberRead; i++) {
        if (buffer[i] == '\"') {
            quote = i;
            break;
        }
    }
    
    if (quote != -1) {
        return new String(buffer, offset, quote - offset);
    }
    else {
        StringBuffer nameBuffer = new StringBuffer();
        nameBuffer.append(buffer, offset, numberRead - offset);
        int remaining = ENCODING_BUFFER_LENGTH - total;
        for (int i = 0; i < remaining; i++) {
            int b = input.read();
            if (b == -1) {
                break;
            }
            else if (b == '\"') {
                break;
            }
            else {
                nameBuffer.append((char)b);
            }
        }
        return nameBuffer.toString();
    }
}

/**
 * If the begining of a title is found, create a new StringBuffer and fill it with as much
 * of the title as is available. If the whole title is found, record it in the instance variable
 * <tt>title</tt> and return null.
 *
 * @return null if the title is not found or the whole title is found. If just the beginning of the title
 * is found, return a StringBuffer with what has been read so far.
 */ 
private StringBuffer createBufferIfTitleBegins(SearchSieve titleFinder, char[] buffer, int numberRead) {
    int openingIndex =
        titleFinder.scanLowerCase(buffer, 0, numberRead);
    StringBuffer titleBuffer = null;
    if (openingIndex != -1) {
        int closingIndex = -1;
        for (int i = openingIndex + 1; i < numberRead; i++) {
            if (buffer[i] == '<') {
                closingIndex = i;
                break;
            }
        }
        if (closingIndex != -1) {
            title = new String(buffer, openingIndex + 1, closingIndex - openingIndex - 1);
            gotTitle = true;
        }
        else {
            titleBuffer = new StringBuffer();
            titleBuffer.append(buffer, 0, numberRead - openingIndex - 1);
        }
    }
    return titleBuffer;
}

/**
 * Append the range of char in buffer from 0 up to (but excluding) the first occurrence of
 * delimeter witin length char.
 *
 * @return true if the delimeter was found, false otherwise.
 */
private boolean appendUpTo(StringBuffer stringBuffer, char[] buffer, int length, char delimeter) {
    int closingIndex;
    for (closingIndex = 0; closingIndex < length; closingIndex++) {
        if (buffer[closingIndex] == '<') {
            break;
        }
    }
    stringBuffer.append(buffer, 0, closingIndex);
    return closingIndex == length;
}

public void beginReadingFor(DocumentScanner delegate)
    throws IOException, SecurityException
{
    char[] buffer = delegate.buffer(this);
    Reader reader = openConnection();
    boolean moreToRead = true;
    
    SearchSieve titleFinder = null;
    StringBuffer titleBuffer = null;
    if (!gotTitle) {
        titleFinder = new SearchSieve("<title>");
    }
    
    while (delegate.shouldContinue(this)) {
        int numberRead = reader.read(buffer);
        if (numberRead == -1) {
            break;
        }
        
        if (!gotTitle) {
            if (titleBuffer == null) {
                titleBuffer = createBufferIfTitleBegins(titleFinder, buffer, numberRead);
                if (titleBuffer != null) {
                    titleFinder = null;
                }
            }
            else if (appendUpTo(titleBuffer, buffer, numberRead, '<')) {
                title = titleBuffer.toString();
                gotTitle = true;
                titleBuffer = null;
            }
        }
        
        delegate.scan(buffer, numberRead, this);
    }
    
    gotTitle = true;
    reader.close();
}


public String toString() {
    return url().toString();
}

public boolean equals(Object o) {
    if (!(o instanceof Document)) {
        return false;
    }
    
    return url.equals(((Document)o).url());
}

public int hashCode() {
    return url.hashCode();
}

}
