package com.revfad.search;

/**
 * I receive chunks of a Document as it reads itself.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 */
public interface DocumentScanner {

/**
 * @return The buffer to be passed to scan.
 * @param document The Document calling this method.
 */
public char[] buffer(Document document);

/**
 * Determine whether the scanning should continue. Even after a
 * call to this method has returned false, the document may
 * yet continue reading itself. However, it will make no more calls
 * to my scan method.
 *
 * @param document The Document calling this method.
 */
public boolean shouldContinue(Document document);

/**
 * @param buffer char read from the document.
 * @param length The number of valid char in <tt>buffer</tt>.
 * @param document The Document calling this method.
 */
public void scan(char[] buffer, int length, Document document);

}