package com.revfad.search;

/**
 * I receive the results of a SearchThread's search.
 * If I call the SearchThread's stopSearching method
 * inside any of the methods below, that thread will
 * not call anymore of my methods and will stop searching,
 * perhaps interrupting a present search.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 *
 * @author David Faden, dfaden@iastate.edu
 */
public interface SearchThreadListener {

//XXX! Should we also have this interface be responsible
//for providing the SearchThread with the next
//document to search?

/**
 * Called when a thread starts searching a document. 
 *
 * @param thread The calling SearchThread.
 * @param document The document being searched by thread.
 */
public void searching(SearchThread thread, Document document);

/**
 * Called when a thread finds a match.
 *
 * @param thread The calling SearchThread.
 * @param match The match that was just found.
 */
public void foundMatch(SearchThread thread, Match match);

/**
 * Called when a thread finishes with a document.
 *
 * @param thread The calling SearchThread.
 * @param document The document that was searched.
 */
public void finishedSearching(SearchThread thread, Document document);

/**
 * Called when a thread encounters a reference to another document
 * as through an href or a frame src.
 *
 * @param thread The calling SearchThread.
 * @param document The document containing the link.
 * @param relativeUrl The text of the reference.
 * @param linkText The text of the link. May be null.
 */
public void foundLink(SearchThread thread, Document document, String relativeUrl, String linkText);

}