package com.revfad.search.tests;

import com.revfad.search.*;

import java.io.*;
import java.net.*;
import java.util.*;
import junit.framework.*;

/**
 * If you have JUnit on your classpath, you can run this test
 * as follows:
 * java junit.swingui.TestRunner com.revfad.search.tests.SearchThreadTest
 */
public class SearchThreadTest extends TestCase {

/**
 * I spawn a single thread and perform a non-spidering search.
 */
private static final class SimpleSearch implements SearchThreadListener {

public SimpleSearch(SearchSieve[] query, Hashtable options, URL[] urls) {
    queue = new DocumentQueue();
    for (int i = 0; i < urls.length; i++) {
        queue.addLast(new Document(urls[i], null));
    }
    thread = new SearchThread(query, queue, this, options);
    matches = new Vector();
    allDocumentsSearched = new Vector();
    links = new Vector();
}

public void searching(SearchThread thread, Document document) {
    System.out.println("Searching: " + document.fileName());
}

public void foundMatch(SearchThread thread, Match match) {
    matches.addElement(match);
}

public void finishedSearching(SearchThread thread, Document document) {
    allDocumentsSearched.addElement(document);
    if (queue.isEmpty()) {
        thread.stopSearching();
        queue.kill();
    }
}

public void foundLink(SearchThread thread, Document document, String relativeUrl, String linkText) {
    try {
        links.add(document.buildUrl(relativeUrl));
    }
    catch (MalformedURLException mue) {
        mue.printStackTrace();
    }
}

public void search() {
    thread.start();
    try {
        thread.join();
    }
    catch (InterruptedException ie) {
        throw new RuntimeException(ie.toString());
    }
}

public Vector matches() {
    return matches;
}

public Vector allDocumentsSearched() {
    return allDocumentsSearched;
}

public Vector links() {
    return links;
}

private DocumentQueue queue;

private SearchThread thread;

private Vector matches;

private Vector allDocumentsSearched;

private Vector links;


}

public static Test suite() {
    return new TestSuite(SearchThreadTest.class);
}

private URL fileNameToLocalUrl(String fileName) {
    return LocalFileNameToUrl.toUrl(fileName);
}

private URL[] fileNamesToLocalUrls(String[] fileNames) {
    URL[] urls = new URL[fileNames.length];
    for (int i = 0; i < urls.length; i++) {
        urls[i] = fileNameToLocalUrl(fileNames[i]);
    }
    return urls;
}  

public void test8859Matches() {
    //Search StrangeTalk
    Hashtable options = new Hashtable();
    options.put(SearchThread.Option.TRAILING_CONTEXT_LENGTH, new Integer(20));
    options.put(SearchThread.Option.LEADING_CONTEXT_LENGTH, new Integer(20));
    options.put(SearchThread.Option.MUST_MATCH_ALL, Boolean.TRUE);
    options.put(SearchThread.Option.IGNORE_TAG_CONTENT, Boolean.TRUE);
    options.put(SearchThread.Option.EXACT_MATCHES_ONLY, Boolean.TRUE);
    SimpleSearch search = new SimpleSearch(SearchSieve.parseQuery("\"litter box\" relationships", true), options, fileNamesToLocalUrls(new String[]{"iso_8859_1.html"}));
    search.search();
    
    HashSet expected = new HashSet();
    expected.add(new Match("relationships", new Document(fileNameToLocalUrl("iso_8859_1.html"), null), null, "top"));
    assertEquals(expected, new HashSet(search.matches()));
}

public void testEnglishMatches() {
    //Search StrangeTalk and Brad's Web Page
    Hashtable options = new Hashtable();
    options.put(SearchThread.Option.TRAILING_CONTEXT_LENGTH, new Integer(10));
    options.put(SearchThread.Option.LEADING_CONTEXT_LENGTH, new Integer(5));
    options.put(SearchThread.Option.IGNORE_TAG_CONTENT, Boolean.TRUE);
    options.put(SearchThread.Option.EXACT_MATCHES_ONLY, Boolean.TRUE);
    SimpleSearch search = new SimpleSearch(SearchSieve.parseQuery("\"litter box\" \"web page\"", true), options, fileNamesToLocalUrls(new String[]{"iso_8859_1.html", "macintosh.html"}));
    search.search();
    
    HashSet expected = new HashSet();
    expected.add(new Match("litter box", new Document(fileNameToLocalUrl("iso_8859_1.html"), null), null, "top"));
    expected.add(new Match("web page", new Document(fileNameToLocalUrl("macintosh.html"), null), null, null));
    assertEquals(expected, new HashSet(search.matches()));

}

public void testChineseMatches() {
    Hashtable chineseOptions = new Hashtable();
    chineseOptions.put(SearchThread.Option.TRAILING_CONTEXT_LENGTH, new Integer(20));
    chineseOptions.put(SearchThread.Option.LEADING_CONTEXT_LENGTH, new Integer(20));
    //chineseOptions.put(SearchThread.Option.ALLOW_MUTLIPLE_MATCHES, Boolean.TRUE);
    SimpleSearch chineseSearch = new SimpleSearch(SearchSieve.parseQuery("\u4e0a", false), chineseOptions,
                        fileNamesToLocalUrls(new String[]{"big5.html", "gb2312.html"}));
    chineseSearch.search();

    Set expected = new HashSet();
    expected.add(new Match("\u4e0a", new Document(fileNameToLocalUrl("big5.html"), null),
        null, null));
    expected.add(new Match("\u4e0a", new Document(fileNameToLocalUrl("gb2312.html"), null),
        null, null));
    System.out.println(chineseSearch.matches());
    assertEquals(expected, new HashSet(chineseSearch.matches()));
}

}